package net.sf.saxon.sort;
import net.sf.saxon.Configuration;
import net.sf.saxon.expr.*;
import net.sf.saxon.trans.DynamicError;
import net.sf.saxon.trans.XPathException;
import net.sf.saxon.value.EmptySequence;
import net.sf.saxon.value.StringValue;

import java.util.Comparator;
import java.util.Properties;

/**
* A FixedSortKeyDefinition is a SortKeyDefinition in which all aspects of the
* sort key definition (sort order, data type, etc,) are known.
* A SortKeyDefinition defines one component of a sort key. <BR>
*
*/


public final class FixedSortKeyDefinition extends SortKeyDefinition {

    public FixedSortKeyDefinition() {}

    private transient Comparator comparer = null;
    // Note, the "collation" defines the collating sequence for the sort key. The
    // "comparer" is what is actually used to do comparisons, after taking into account
    // ascending/descending, caseOrder, etc.

    public SortKeyDefinition simplify() {
        return this;
    }

    /**
    * Eliminate dependencies of the sort key definition on the context. For the sort key select
    * expression, this means things that don't depend on the individual node: specifically, variables
    * and current-group(). For the AVTs used to select data type, case order, language, it means
    * all dependencies: after reduction, these values will be constants.
    */

    public FixedSortKeyDefinition reduce(XPathContext context) {
            return this;
    }

    /**
    * Allocate a reusable Comparer to implement this sort key comparison
    */

    public void bindComparer(XPathContext context) throws XPathException {

        String orderX = ((StringValue)order).getStringValue();
        String caseOrderX = ((StringValue)caseOrder).getStringValue();
        String languageX = ((StringValue)language).getStringValue();

        final Configuration config = context.getConfiguration();

        Comparator comp;
        if (collation != null) {
            comp = collation;
        } else {
            Properties props = new Properties();
            if (!languageX.equals("")) {
                props.setProperty("lang", languageX);
            }
            if (!caseOrderX.equals("#default")) {
                props.setProperty("case-order", caseOrderX);
            }
            comp = config.getPlatform().makeCollation(config, props);
        }

        if (dataTypeExpression==null || dataTypeExpression instanceof EmptySequence) {
            if (isBackwardsCompatible()) {
                sortKey = new FirstItemExpression(sortKey);
            } else {
                RoleLocator role =
                    new RoleLocator(RoleLocator.INSTRUCTION, "xsl:sort/select", 0, null);
                role.setErrorCode("XTTE1020");
                sortKey = CardinalityChecker.makeCardinalityChecker(sortKey, StaticProperty.ALLOWS_ZERO_OR_ONE, role);
            }
            comp = new AtomicSortComparer(comp, context);
            ((AtomicSortComparer)comp).setEmptyLeast(emptyLeast);
        } else {
            String dataType = ((StringValue)dataTypeExpression).getStringValue();
            if (dataType.equals("text")) {
                comp = new TextComparer(comp);
            } else if (dataType.equals("number")) {
                comp = new NumericComparer();
            } else {
                DynamicError err = new DynamicError("data-type on xsl:sort must be 'text' or 'number'");
                err.setErrorCode("XTDE0030");
                throw err;
            }
        }

        if (orderX.equals("ascending")) {
            comparer = comp;
        } else if (orderX.equals("descending")) {
            comparer = new DescendingComparer(comp);
        } else {
            DynamicError err1 = new DynamicError("order must be 'ascending' or 'descending'");
            err1.setErrorCode("XTDE0030");
            throw err1;
        }

        if (stable != null) {
            StringValue stableVal = (StringValue)stable.evaluateItem(context);
            String s = stableVal.getStringValue().trim();
            if (s.equals("yes") || s.equals("no")) {
                // no action
            } else {
                DynamicError err = new DynamicError("Value of 'stable' on xsl:sort must be 'yes' or 'no'");
                err.setErrorCode("XTDE0030");
                throw err;
            }
        }
    }

    /**
    * Get the comparer which is used to compare two values according to this sort key.
    */

    public Comparator getComparer() {
        return comparer;
    }

}


//
// The contents of this file are subject to the Mozilla Public License Version 1.0 (the "License");
// you may not use this file except in compliance with the License. You may obtain a copy of the
// License at http://www.mozilla.org/MPL/
//
// Software distributed under the License is distributed on an "AS IS" basis,
// WITHOUT WARRANTY OF ANY KIND, either express or implied.
// See the License for the specific language governing rights and limitations under the License.
//
// The Original Code is: all this file.
//
// The Initial Developer of the Original Code is Michael H. Kay.
//
// Portions created by (your name) are Copyright (C) (your legal entity). All Rights Reserved.
//
// Contributor(s): none.
//
